/*
 * lifstat.c
 * fsf - 2003-08-01
 * last modified: 2003-08-23
 *
 * Copyright (c) 2003, Frank S. Fejes III
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  * Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "lifstat.h"

void usage_display(const char *name)
{
    printf("usage: %s [-vhndt] [-i interface] [interval] [count]\n", name);
}

void version_display(void)
{
    printf("lifstat %s - written by Frank S. Fejes III\n", VERSION);
}

void header_display(int show_date, int show_time)
{
    if (show_date) { printf("      "); }
    if (show_time) { printf("         "); }
    printf("%-5s %9s %9s %5s %9s\n", "", "in", "", "", "out");
    if (show_date) { printf("date  "); }
    if (show_time) { printf("time     "); }
    printf("%-5s %9s %9s %5s ", "iface", "bytes", "pkts", "err");
    printf("%9s %9s %5s %5s\n", "bytes", "pkts", "err", "coll");
}

void stats_display(const char *ifn, stat_struct *cs, stat_struct *ps, int n,
    const char *ts)
{
    int i;
    char *iface;
    long long in_bytes, in_pkts;
    unsigned long in_errs;
    long long out_bytes, out_pkts;
    unsigned long out_errs;
    unsigned long coll;

    for (i = 0; i < n; ++i) {
        if (strcmp(cs[i].iface, ps[i].iface) != 0) {
            continue;
        }

        iface = cs[i].iface;
        /* if the interface name is NULL, display all */
        /* if not NULL, only display that interface   */
        if (ifn && (strcmp(ifn, iface) != 0)) {
            continue;
        }

        in_bytes = cs[i].in_bytes - ps[i].in_bytes;
        in_pkts = cs[i].in_pkts - ps[i].in_pkts;
        in_errs = cs[i].in_errs - ps[i].in_errs;
        out_bytes = cs[i].out_bytes - ps[i].out_bytes;
        out_pkts = cs[i].out_pkts - ps[i].out_pkts;
        out_errs = cs[i].out_errs - ps[i].out_errs;
        coll = cs[i].coll - ps[i].coll;

        if (ts) {
            printf("%s %-5s %9lld %9lld %5lu ", ts,
                iface, in_bytes, in_pkts, in_errs);
        }
        else {
            printf("%-5s %9lld %9lld %5lu ", iface, in_bytes, in_pkts, in_errs);
        }
        printf("%9lld %9lld %5lu %5lu\n", out_bytes, out_pkts, out_errs, coll);
    }
}

int read_line(int fd, char *line)
{
    char buf[1];
    int n = 0;

    while (read(fd, buf, 1) == 1) {
        if (buf[0] == '\n') {
            line[n] = 0;
            break;
        }
        line[n++] = buf[0];
    }

    line[n] = 0;

    return(n);
}

int line_parse(char *line, int len, stat_struct *stats)
{
    char *ptr = line;
    char *index;
    int i, n;
    char *cols[MAX_COLS];

    /* cut out leading spaces */
    for (i = 0; (i < len) && (line[i] == ' '); ++i, ++ptr);

    /* grab the interface name */
    index = strchr(ptr, ':');
    if (index) {
        *index = 0;
        strcpy(stats->iface, ptr);
        ptr = index + 1;
    }

    for (i = 0; i < MAX_COLS; ++i) {
        cols[i] = NULL;
    }

    n = split(ptr, cols, MAX_COLS, ' ');
    if (n < DATA_COLS) {
        fprintf(stderr, "Bad # of columns (%d instead of %d)!\n", n, DATA_COLS);
        return(-1);
    }

    stats->in_bytes = atoll(cols[0]);
    stats->in_pkts = atoll(cols[1]);
    stats->in_errs = strtoul(cols[3], (char **) NULL, 10);
    stats->out_bytes = atoll(cols[8]);
    stats->out_pkts = atoll(cols[9]);
    stats->out_errs = strtoul(cols[10], (char **) NULL, 10);
    stats->coll = strtoul(cols[13], (char **) NULL, 10);

    for (i = 0; i < MAX_COLS; ++i) {
        if (cols[i]) {
            free(cols[i]);
            cols[i] = NULL;
        }
    }

    return(0);
}

int stats_read(stat_struct *stats)
{
    int ifaces = 0;
    char line[MAX_LINE];
    char msg[256];
    FILE *f;

    f = fopen(DEVFILE, "r");
    if (! f) {
        sprintf(msg, "Cannot open file %s", DEVFILE);
        perror(msg);
    }

    /* throw out the first two lines */
    fgets(line, MAX_LINE, f);
    fgets(line, MAX_LINE, f);

    while (fgets(line, MAX_LINE, f) != NULL) {
        if (line_parse(line, strlen(line), &stats[ifaces]) != -1) {
            ++ifaces;
        }
    }

    if (ferror(f)) {
        sprintf(msg, "Error reading %s", DEVFILE);
        perror(msg);
    }

    fclose(f);

    return(ifaces);
}

void loop_display(const char *interface, int interval, int count,
    int show_date, int show_time)
{
    stat_struct prev[MAX_IF];
    stat_struct curr[MAX_IF];
    int i = 0;
    int n;
    char *ts = NULL;
    time_t t;

    if (show_date || show_time) {
        ts = (char *) malloc(32 * sizeof(char));
    }

    if (count <= 0) {
        i = -1;
        count = 0;
    }

    n = stats_read(prev);

    while (i < count) {
        sleep(interval);

        if (ts) {
            t = time(NULL);
            if (show_date && show_time) {
                strftime(ts, 31, "%m-%d %H:%M:%S", localtime(&t));
            }
            else if (show_date) {
                strftime(ts, 31, "%m-%d", localtime(&t));
            }
            else {
                strftime(ts, 31, "%H:%M:%S", localtime(&t));
            }
        }

        n = stats_read(curr);

        stats_display(interface, curr, prev, n, ts);

        memcpy(prev, curr, sizeof(prev));

        if (i > -1) {
            ++i;
        }
    }
}

int main(int argc, char **argv)
{
    int interval = DEF_INTERVAL;
    int count = DEF_COUNT;
    int show_header = DEF_SHOW_HEADER;
    int show_time = DEF_SHOW_TIME;
    int show_date = DEF_SHOW_DATE;
    char *interface = NULL;
    int ch;
    extern int optind;
    extern char *optarg;
 
    while ( (ch = getopt(argc, argv, "i:hvdtn")) != -1) {
        switch(ch) {
        case 'i':
            interface = (char *) malloc(strlen(optarg) + 1);
            if (interface) {
                strcpy(interface, optarg);
            }
            break;
        case 'n':
            show_header = 0;
            break;
        case 'd':
            show_date = 1;
            break;
        case 't':
            show_time = 1;
            break;
        case 'h':
            usage_display(argv[0]);
            exit(0);
            break;
        case 'v':
            version_display();
            exit(0);
            break;
        }
    }
    argc -= optind;
    argv += optind; 

    if (argc) {
        interval = atoi(argv[0]);
    }
    if (argc > 1) {
        count = atoi(argv[1]);
    }

    if (show_header) {
        header_display(show_date, show_time);
    }
    loop_display(interface, interval, count, show_date, show_time);

    exit(0);
}

